"""This module contains the type definition for the Groq call keyword arguments."""

from collections.abc import Sequence

from groq.types.chat import ChatCompletionMessageParam, ChatCompletionToolParam

from ..base import BaseCallKwargs
from .call_params import GroqCallParams


class GroqCallKwargs(GroqCallParams, BaseCallKwargs[ChatCompletionToolParam]):
    model: str
    messages: Sequence[ChatCompletionMessageParam]



"""Tests the `gemini._utils.setup_call` module."""

from unittest.mock import MagicMock, patch

import pytest
from google.generativeai import GenerativeModel  # type: ignore
from google.generativeai.types import GenerationConfig
from google.generativeai.types.content_types import ToolConfigDict
from pydantic import BaseModel

from mirascope.core.gemini._utils._convert_common_call_params import (
    convert_common_call_params,
)
from mirascope.core.gemini._utils._setup_call import setup_call
from mirascope.core.gemini.tool import GeminiTool


@pytest.fixture()
def mock_base_setup_call() -> MagicMock:
    """Returns the mock setup_call function."""
    mock_setup_call = MagicMock()
    mock_setup_call.return_value = [MagicMock() for _ in range(3)] + [{}]
    return mock_setup_call


@patch(
    "mirascope.core.gemini._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.gemini._utils._setup_call._utils", new_callable=MagicMock)
@patch(
    "mirascope.core.gemini._utils._setup_call.GenerativeModel", new_callable=MagicMock
)
@patch(
    "mirascope.core.gemini._utils._setup_call.GenerativeModel.generate_content",
    new_callable=MagicMock,
)
def test_setup_call(
    mock_generate_content: MagicMock,
    mock_generative_model: MagicMock,
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function."""
    generative_model = GenerativeModel(model_name="gemini-1.5-flash")
    mock_generative_model.return_value = generative_model
    mock_utils.setup_call = mock_base_setup_call
    fn = MagicMock()
    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="gemini-1.5-flash",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert prompt_template == mock_base_setup_call.return_value[0]
    assert tool_types == mock_base_setup_call.return_value[2]
    assert "contents" in call_kwargs and call_kwargs["contents"] == messages
    mock_base_setup_call.assert_called_once_with(
        fn, {}, None, None, GeminiTool, {}, convert_common_call_params
    )
    mock_convert_message_params.assert_called_once_with(
        mock_base_setup_call.return_value[1]
    )
    assert messages == mock_convert_message_params.return_value
    mock_generative_model.assert_called_once_with(model_name="gemini-1.5-flash")
    assert create(**call_kwargs)
    mock_generate_content.assert_called_once_with(**call_kwargs)
    mock_generate_content.reset_mock()
    assert create(stream=True, **call_kwargs)
    mock_generate_content.assert_called_once_with(**call_kwargs, stream=True)
    mock_generate_content.reset_mock()
    assert create(stream=False, **call_kwargs)
    mock_generate_content.assert_called_once_with(**call_kwargs)


@pytest.mark.parametrize("generation_config_type", [dict, GenerationConfig])
@patch(
    "mirascope.core.gemini._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.gemini._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_json_mode(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
    generation_config_type: type,
) -> None:
    """Tests the `setup_call` function with JSON mode."""
    mock_utils.setup_call = mock_base_setup_call
    mock_utils.json_mode_content = MagicMock()
    mock_base_setup_call.return_value[1] = [
        {"role": "user", "parts": [{"type": "text", "text": "test"}]}
    ]
    mock_base_setup_call.return_value[-1]["tools"] = MagicMock()
    mock_base_setup_call.return_value[-1]["generation_config"] = generation_config_type(
        candidate_count=1,
        max_output_tokens=100,
        response_mime_type="application/xml",
        response_schema=None,
        stop_sequences=["\n"],
        temperature=0.5,
        top_k=0,
        top_p=0,
    )
    mock_convert_message_params.side_effect = lambda x: x
    _, _, messages, _, call_kwargs = setup_call(
        model="gemini-1.5-flash",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert messages[-1]["parts"][-1] == mock_utils.json_mode_content.return_value
    assert "tools" in call_kwargs
    assert "generation_config" in call_kwargs
    assert call_kwargs["generation_config"] == {
        "candidate_count": 1,
        "max_output_tokens": 100,
        "response_mime_type": "application/json",
        "response_schema": None,
        "stop_sequences": ["\n"],
        "temperature": 0.5,
        "top_k": 0,
        "top_p": 0,
    }


@pytest.mark.parametrize("generation_config_type", [dict, GenerationConfig])
@patch(
    "mirascope.core.gemini._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.gemini._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_system_instruction(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
    generation_config_type: type,
) -> None:
    """Tests the `setup_call` function with JSON mode."""
    mock_utils.setup_call = mock_base_setup_call
    mock_utils.json_mode_content = MagicMock()
    mock_base_setup_call.return_value[1] = [
        {
            "role": "system",
            "parts": ["system_instruction test"],
        },
        {"role": "user", "parts": [{"type": "text", "text": "test"}]},
    ]
    mock_client = MagicMock()
    mock_client._system_instruction = None
    mock_base_setup_call.return_value[-1]["tools"] = MagicMock()
    mock_base_setup_call.return_value[-1]["generation_config"] = generation_config_type(
        candidate_count=1,
        max_output_tokens=100,
        response_mime_type="application/xml",
        response_schema=None,
        stop_sequences=["\n"],
        temperature=0.5,
        top_k=0,
        top_p=0,
    )
    mock_convert_message_params.side_effect = lambda x: x
    _, _, messages, _, call_kwargs = setup_call(
        model="gemini-1.5-flash",
        client=mock_client,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert messages[-1]["parts"][-1] == mock_utils.json_mode_content.return_value
    assert mock_client._system_instruction.parts[0].text == "system_instruction test"


@patch(
    "mirascope.core.gemini._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.gemini._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_extract(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function with extraction."""
    mock_tool = MagicMock()
    mock_tool._name.side_effect = lambda: "test"
    mock_base_setup_call.return_value[2] = [mock_tool]
    mock_utils.setup_call = mock_base_setup_call
    _, _, _, _, call_kwargs = setup_call(
        model="gemini-1.5-flash",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=BaseModel,
        stream=False,
    )
    assert "tool_config" in call_kwargs and isinstance(
        call_kwargs["tool_config"], ToolConfigDict
    )
    assert call_kwargs["tool_config"].function_calling_config == {
        "allowed_function_names": ["test"],
        "mode": "any",
    }



import base64
import json
import re
from typing import cast

from mistralai.models import (
    AssistantMessage,
    ImageURLChunk,
    ReferenceChunk,
    SystemMessage,
    TextChunk,
    ToolMessage,
    UserMessage,
)

from mirascope.core.base._utils._base_message_param_converter import (
    BaseMessageParamConverter,
)
from mirascope.core.mistral._utils import convert_message_params

from ...base import BaseMessageParam, ImagePart, TextPart, ToolResultPart
from ...base.message_param import ImageURLPart, ToolCallPart


def _create_image_part_from_data_url(image_url: str) -> ImagePart | None:
    match = re.match(r"data:(image/\w+);base64,(.+)", image_url)
    if not match:
        return None
    mime_type = match.group(1)
    image_base64 = match.group(2)
    image_data = base64.b64decode(image_base64)
    return ImagePart(
        type="image",
        media_type=mime_type,
        image=image_data,
        detail=None,
    )


class MistralMessageParamConverter(BaseMessageParamConverter):
    """Converts between Mistral `AssistantMessage` and Mirascope `BaseMessageParam`."""

    @staticmethod
    def to_provider(message_params: list[BaseMessageParam]) -> list:
        """
        Convert from Mirascope `BaseMessageParam` to Mistral messages (AssistantMessage, etc.).
        Mistral’s code snippet references convert_message_params returning a list of
        [AssistantMessage|SystemMessage|ToolMessage|UserMessage].
        """
        return convert_message_params(
            cast(
                list[
                    BaseMessageParam
                    | AssistantMessage
                    | SystemMessage
                    | ToolMessage
                    | UserMessage
                ],
                message_params,
            )
        )

    @staticmethod
    def from_provider(
        message_params: list[
            AssistantMessage | ToolMessage | SystemMessage | UserMessage
        ],
    ) -> list[BaseMessageParam]:
        """
        Convert from Mistral's `AssistantMessage` to Mirascope `BaseMessageParam`.
        """
        converted: list[BaseMessageParam] = []
        for message_param in message_params:
            content = message_param.content

            converted_parts = []

            if tool_calls := getattr(message_param, "tool_calls", None):
                for tool in tool_calls:
                    arguments = tool.function.arguments
                    converted.append(
                        BaseMessageParam(
                            role="tool",
                            content=[
                                ToolCallPart(
                                    type="tool_call",
                                    name=tool.function.name,
                                    id=tool.id,
                                    args=json.loads(arguments)
                                    if isinstance(arguments, str)
                                    else arguments,
                                )
                            ],
                        )
                    )
            elif isinstance(message_param, ToolMessage):
                converted.append(
                    BaseMessageParam(
                        role="tool",
                        content=[
                            ToolResultPart(
                                type="tool_result",
                                name=message_param.name,  # pyright: ignore [reportArgumentType]
                                content=message_param.content,  # pyright: ignore [reportArgumentType]
                                id=message_param.tool_call_id,  # pyright: ignore [reportArgumentType]
                            )
                        ],
                    )
                )
            elif isinstance(content, str):
                converted_parts.append(TextPart(type="text", text=content))
            elif isinstance(content, list):
                for chunk in content:
                    if isinstance(chunk, TextChunk):
                        converted_parts.append(TextPart(type="text", text=chunk.text))

                    elif isinstance(chunk, ImageURLChunk):
                        image_url = chunk.image_url
                        if isinstance(image_url, str):
                            if image_part := _create_image_part_from_data_url(
                                image_url
                            ):
                                converted_parts.append(image_part)
                            else:
                                converted_parts.append(
                                    ImageURLPart(
                                        type="image_url",
                                        url=image_url,
                                        detail=None,
                                    )
                                )

                        else:
                            img_url_str = image_url.url  # type: ignore
                            if image_part := _create_image_part_from_data_url(
                                img_url_str
                            ):
                                converted_parts.append(image_part)
                            else:
                                converted_parts.append(
                                    ImageURLPart(
                                        type="image_url",
                                        url=img_url_str,
                                        detail=image_url.detail
                                        if isinstance(image_url.detail, str)
                                        else None,
                                    )
                                )
                    elif isinstance(chunk, ReferenceChunk):
                        raise ValueError(
                            "ReferenceChunk is not supported for conversion to BaseMessageParam."
                        )

                    else:
                        # Unknown chunk type
                        raise ValueError(
                            f"Unsupported ContentChunk type: {type(chunk).__name__}"
                        )

            if len(converted_parts) == 1 and isinstance(converted_parts[0], TextPart):
                converted.append(
                    BaseMessageParam(
                        role=message_param.role,  # pyright: ignore [reportArgumentType]
                        content=converted_parts[0].text,
                    )
                )
            else:
                if converted_parts:
                    converted.append(
                        BaseMessageParam(
                            role=message_param.role,  # pyright: ignore [reportArgumentType]
                            content=converted_parts,
                        )
                    )

        return converted



"""Tests the `call_response` module."""

import base64
import json
from collections.abc import Sequence
from typing import Any
from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseModel

from mirascope.core.base._utils import BaseMessageParamConverter
from mirascope.core.base.call_response import BaseCallResponse, transform_tool_outputs
from mirascope.core.base.message_param import BaseMessageParam


def test_base_call_response() -> None:
    """Tests the `BaseCallResponse` class."""

    class MyMessageParamConverter(BaseMessageParamConverter):
        @staticmethod
        def from_provider(message_params: list[dict]) -> list[BaseMessageParam]:
            """Converts provider-specific messages -> Base message params."""
            return [
                BaseMessageParam(
                    role=message_param["role"], content=message_param["content"]
                )
                for message_param in message_params
            ]

    class MyCallResponse(BaseCallResponse):
        _message_converter: type[MyMessageParamConverter] = MyMessageParamConverter

        @property
        def content(self) -> str:
            return "content"

    patch.multiple(MyCallResponse, __abstractmethods__=set()).start()
    call_response = MyCallResponse(
        metadata={},
        response="",
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[{"role": "user", "content": "content"}],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )  # type: ignore
    assert str(call_response) == "content"
    tool = MagicMock()
    tool._name = lambda: "mock_tool"
    assert call_response.serialize_tool_types([tool], info=MagicMock()) == [
        {"type": "function", "name": "mock_tool"}
    ]
    assert call_response.common_tools is None
    assert call_response.common_usage is None
    assert call_response.common_messages == [
        BaseMessageParam(role="user", content="content")
    ]


def test_base_call_response_without_model() -> None:
    """Tests the `BaseCallResponse` class without model."""

    class MyCallResponse(BaseCallResponse):
        @property
        def content(self) -> str:
            return "content"

    patch.multiple(MyCallResponse, __abstractmethods__=set()).start()
    call_response = MyCallResponse(
        metadata={},
        response="",
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
        _model=None,  # pyright: ignore [reportArgumentType]
    )  # type: ignore
    assert str(call_response) == "content"
    tool = MagicMock()
    tool._name = lambda: "mock_tool"
    assert call_response.serialize_tool_types([tool], info=MagicMock()) == [
        {"type": "function", "name": "mock_tool"}
    ]
    assert call_response.common_tools is None
    assert call_response.common_usage is None
    assert call_response.cost is None


class SimpleModel(BaseModel):
    name: str
    value: int


class NestedModel(BaseModel):
    simple: SimpleModel
    values: list[int]


@pytest.fixture
def process_tools():
    @transform_tool_outputs
    def _process_tools(
        cls: Any, tools_and_outputs: Sequence[tuple[Any, str]]
    ) -> list[str]:
        return [output for _, output in tools_and_outputs]

    return _process_tools


@pytest.mark.parametrize(
    "input_value,expected",
    [
        ("string value", "string value"),
        (42, "42"),
        (3.14, "3.14"),
        (True, "true"),
        (b"binary data", base64.b64encode(b"binary data").decode("utf-8")),
    ],
)
def test_transform_tool_outputs_primitive_types(
    process_tools: Any, input_value: Any, expected: str
) -> None:
    tool = MagicMock()
    result = process_tools(None, [(tool, input_value)])
    assert result[0] == expected


@pytest.mark.parametrize(
    "input_model,expected_dict",
    [
        (SimpleModel(name="test", value=1), {"name": "test", "value": 1}),
        (
            NestedModel(simple=SimpleModel(name="nested", value=2), values=[1, 2, 3]),
            {"simple": {"name": "nested", "value": 2}, "values": [1, 2, 3]},
        ),
    ],
)
def test_transform_tool_outputs_pydantic_models(
    process_tools: Any, input_model: BaseModel, expected_dict: dict
) -> None:
    tool = MagicMock()
    result = process_tools(None, [(tool, input_model)])
    assert json.loads(result[0]) == expected_dict


@pytest.mark.parametrize(
    "container_type,input_value,expected_type",
    [
        ("list", ["string", 42, SimpleModel(name="item", value=3)], list),
        (
            "tuple",
            ("string", 42, SimpleModel(name="item", value=3)),
            list,
        ),
    ],
)
def test_transform_tool_outputs_containers(
    process_tools: Any, container_type: str, input_value: Any, expected_type: type
) -> None:
    tool = MagicMock()
    result = process_tools(None, [(tool, input_value)])
    parsed = json.loads(result[0])
    assert isinstance(parsed, expected_type)
    assert parsed == ["string", 42, {"name": "item", "value": 3}]


def test_transform_tool_outputs_sets(process_tools: Any) -> None:
    tool = MagicMock()
    result = process_tools(None, [(tool, {"string", 42, 3.14})])
    parsed = json.loads(result[0])
    assert isinstance(parsed, list)
    assert set(parsed) == {"string", 42, 3.14}


def test_transform_tool_outputs_nested_dictionaries(process_tools: Any) -> None:
    """Tests the serialization of nested dictionaries."""

    class NestedModel(BaseModel):
        simple: SimpleModel
        values: list[int]

    tool = MagicMock()
    result = process_tools(
        None,
        [(tool, {"simple": SimpleModel(name="test", value=1), "values": [1, 2, 3]})],
    )
    parsed = json.loads(result[0])
    assert parsed == {"simple": {"name": "test", "value": 1}, "values": [1, 2, 3]}


def test_transform_tool_outputs_unsupported_type(process_tools: Any) -> None:
    """Tests the serialization of an unsupported type."""

    class UnsupportedType:
        pass

    tool = MagicMock()
    with pytest.raises(TypeError) as exc_info:
        process_tools(None, [(tool, UnsupportedType())])
    assert "Unsupported type for serialization" in str(exc_info.value)



"""The `LiteLLMStream` class for convenience around streaming LLM calls.

usage docs: learn/streams.md
"""

from collections.abc import AsyncGenerator, Generator

from litellm import Choices, Message
from litellm.cost_calculator import completion_cost
from litellm.types.utils import ModelResponse

from ..base.types import CostMetadata
from ..openai import OpenAIStream, OpenAITool
from .call_response import LiteLLMCallResponse
from .call_response_chunk import LiteLLMCallResponseChunk


class LiteLLMStream(OpenAIStream):
    """A simple wrapper around `OpenAIStream`.

    Everything is the same except updates to the `construct_call_response` method and
    the `cost` property so that cost is properly calculated using LiteLLM's cost
    calculation method. This ensures cost calculation works for non-OpenAI models.
    """

    _provider = "litellm"

    def __iter__(
        self,
    ) -> Generator[tuple[LiteLLMCallResponseChunk, OpenAITool | None], None, None]:
        yield from super().__iter__()  # pyright: ignore [reportReturnType]

    def __aiter__(
        self,
    ) -> AsyncGenerator[tuple[LiteLLMCallResponseChunk, OpenAITool | None], None]:
        return super().__aiter__()  # pyright: ignore [reportReturnType] # pragma: no cover

    @property
    def cost_metadata(self) -> CostMetadata:
        """Returns metadata needed for cost calculation."""
        response = self.construct_call_response()
        return CostMetadata(
            cost=response.cost,
        )

    def construct_call_response(self) -> LiteLLMCallResponse:
        openai_call_response = super().construct_call_response()
        openai_response = openai_call_response.response
        model_response = ModelResponse(
            id=openai_response.id,
            choices=[
                Choices(
                    finish_reason=choice.finish_reason,
                    index=choice.index,
                    message=Message(**choice.message.model_dump()),
                    logprobs=choice.logprobs,
                )
                for choice in openai_response.choices
            ],
            created=openai_response.created,
            model=openai_response.model,
            object=openai_response.object,
            system_fingerprint=openai_response.system_fingerprint,
            usage=openai_response.usage.model_dump() if openai_response.usage else None,
        )
        model_response._hidden_params["response_cost"] = completion_cost(
            model=self.model,
            messages=openai_call_response.messages,
            completion=openai_call_response.content,
        )
        response = LiteLLMCallResponse(
            metadata=openai_call_response.metadata,
            response=model_response,  # pyright: ignore [reportArgumentType]
            tool_types=openai_call_response.tool_types,
            prompt_template=openai_call_response.prompt_template,
            fn_args=openai_call_response.fn_args,
            dynamic_config=openai_call_response.dynamic_config,
            messages=openai_call_response.messages,
            call_params=openai_call_response.call_params,
            call_kwargs=openai_call_response.call_kwargs,
            user_message_param=openai_call_response.user_message_param,
            start_time=openai_call_response.start_time,
            end_time=openai_call_response.end_time,
        )
        response._model = self.model
        return response



"""Tests the `groq.call_response_chunk` module."""

from groq.types.chat import ChatCompletionChunk
from groq.types.chat.chat_completion_chunk import (
    Choice,
    ChoiceDelta,
    ChoiceDeltaToolCall,
    ChoiceDeltaToolCallFunction,
)
from groq.types.completion_usage import CompletionUsage

from mirascope.core.base.types import CostMetadata
from mirascope.core.groq.call_response_chunk import GroqCallResponseChunk


def test_groq_call_response_chunk() -> None:
    """Tests the `GroqCallResponseChunk` class."""
    tool_call = ChoiceDeltaToolCall(
        index=0,
        id="id",
        function=ChoiceDeltaToolCallFunction(
            arguments='{"key": "value"}', name="function"
        ),
        type="function",
    )
    choices = [
        Choice(
            delta=ChoiceDelta(content="content", tool_calls=[tool_call]),
            index=0,
            finish_reason="stop",
        )
    ]
    usage = CompletionUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2)
    chunk = ChatCompletionChunk(
        id="id",
        choices=choices,
        created=0,
        model="llama3-70b-8192",
        object="chat.completion.chunk",
        usage=usage,
        x_groq=None,
    )
    call_response_chunk = GroqCallResponseChunk(chunk=chunk)
    assert call_response_chunk.content == "content"
    assert call_response_chunk.finish_reasons == ["stop"]
    assert call_response_chunk.model == "llama3-70b-8192"
    assert call_response_chunk.id == "id"
    assert call_response_chunk.usage == usage
    assert call_response_chunk.input_tokens == 1
    assert call_response_chunk.output_tokens == 1
    assert call_response_chunk.common_finish_reasons == ["stop"]
    assert call_response_chunk.cost_metadata == CostMetadata(
        input_tokens=1,
        output_tokens=1,
        cached_tokens=0,
    )


def test_groq_call_response_chunk_no_choices_or_usage() -> None:
    """Tests the `GroqCallResponseChunk` class with None values."""
    chunk = ChatCompletionChunk(
        id="id",
        choices=[],
        created=0,
        model="llama3-70b-8192",
        object="chat.completion.chunk",
        usage=None,
        x_groq=None,
    )
    call_response_chunk = GroqCallResponseChunk(chunk=chunk)
    assert call_response_chunk.content == ""
    assert call_response_chunk.usage is None
    assert call_response_chunk.input_tokens is None
    assert call_response_chunk.output_tokens is None
    assert call_response_chunk.common_finish_reasons == []



from collections import namedtuple
from unittest.mock import MagicMock, patch

import pytest
from docker.errors import APIError, ImageNotFound

from mirascope.tools import DockerOperationToolKit, DockerOperationToolKitConfig
from mirascope.tools.system._docker_operation import DockerContainer

# Create ExecResult namedtuple to match docker's implementation
ExecResult = namedtuple("ExecResult", "exit_code,output")


@pytest.fixture
def mock_docker_client():
    """Creates a mock Docker client."""
    with patch("docker.from_env") as mock_client:
        # Mock container and its methods
        mock_container = MagicMock()
        mock_container.put_archive.return_value = True
        mock_container.stop.return_value = None

        # Initial exec result with the correct namedtuple format
        mock_container.exec_run.return_value = ExecResult(0, b"Test output")

        # Configure client to return mock container
        mock_client.return_value.containers.run.return_value = mock_container
        yield mock_client


@pytest.fixture
def docker_operation_toolkit(mock_docker_client):
    """Creates a DockerOperationToolKit instance with mocked Docker client."""
    config = DockerOperationToolKitConfig(
        docker_image="python:3.13-slim",
        max_memory="256m",
        allow_network=False,
    )
    return DockerOperationToolKit(
        config=config, docker_container=DockerContainer(config=config)
    )


def test_execute_python_without_custom_config(mock_docker_client):
    """Test executing Python code without custom configuration."""
    docker_operation_toolkit = DockerOperationToolKit(
        docker_container=DockerContainer()
    )
    code = """
print('Hello from Python!')
x = 1 + 1
print(f'Result: {x}')
"""
    tool, _ = docker_operation_toolkit.create_tools()
    assert issubclass(tool, DockerOperationToolKit.ExecutePython)

    # Configure mock response
    mock_container = mock_docker_client.return_value.containers.run.return_value
    mock_container.exec_run.return_value = ExecResult(
        0, b"Hello from Python!\nResult: 2\n"
    )

    result = tool(code=code).call()  # pyright: ignore [reportCallIssue]
    assert "Hello from Python!" in result
    assert "Result: 2" in result


def test_execute_python(
    docker_operation_toolkit: DockerOperationToolKit, mock_docker_client
):
    """Test executing Python code."""
    code = """
print('Hello from Python!')
x = 1 + 1
print(f'Result: {x}')
"""
    tool, _ = docker_operation_toolkit.create_tools()
    assert issubclass(tool, DockerOperationToolKit.ExecutePython)

    # Configure mock response
    mock_container = mock_docker_client.return_value.containers.run.return_value
    mock_container.exec_run.return_value = ExecResult(
        0, b"Hello from Python!\nResult: 2\n"
    )

    result = tool(code=code, requirements=None).call()  # pyright: ignore [reportCallIssue]
    assert "Hello from Python!" in result
    assert "Result: 2" in result


def test_execute_python_with_requirements(mock_docker_client):
    """Test executing Python code with requirements."""
    code = """
import mirascope
print(mirascope)
"""
    config = DockerOperationToolKitConfig(
        docker_image="python:3.13-slim",
        max_memory="256m",
        allow_network=True,
    )
    docker_operation_toolkit = DockerOperationToolKit(
        config=config, docker_container=DockerContainer(config=config)
    )
    tool, _ = docker_operation_toolkit.create_tools()
    assert issubclass(tool, DockerOperationToolKit.ExecutePython)

    # Configure mock responses
    mock_container = mock_docker_client.return_value.containers.run.return_value
    mock_container.exec_run.side_effect = [
        ExecResult(0, b"Successfully installed mirascope"),
        ExecResult(0, b"<module 'mirascope' from '/path/to/mirascope'>"),
    ]

    result = tool(code=code, requirements=["mirascope"]).call()  # pyright: ignore [reportCallIssue]
    assert result.startswith("<module 'mirascope' from")


def test_execute_python_with_requirements_network_disabled(
    docker_operation_toolkit: DockerOperationToolKit,
):
    """Test executing Python code with requirements when network is disabled."""
    code = ""
    tool, _ = docker_operation_toolkit.create_tools()
    assert issubclass(tool, DockerOperationToolKit.ExecutePython)
    result = tool(code=code, requirements=["mirascope"]).call()  # pyright: ignore [reportCallIssue]
    assert (
        result
        == "Error: Network access is disabled. Cannot install requirements via pip."
    )


def test_execute_shell(
    docker_operation_toolkit: DockerOperationToolKit, mock_docker_client
):
    """Test executing shell commands."""
    _, tool = docker_operation_toolkit.create_tools()
    assert issubclass(tool, DockerOperationToolKit.ExecuteShell)

    # Configure mock response
    mock_container = mock_docker_client.return_value.containers.run.return_value
    mock_container.exec_run.return_value = ExecResult(0, b"Hello from shell!\n")

    result = tool(command="echo 'Hello from shell!'").call()  # pyright: ignore [reportCallIssue]
    assert "Hello from shell!" in result


def test_execute_python_with_syntax_error(
    docker_operation_toolkit: DockerOperationToolKit, mock_docker_client
):
    """Test executing Python code with syntax error."""
    code = """
print('Hello'
x = 1 + 
"""
    tool, _ = docker_operation_toolkit.create_tools()
    assert issubclass(tool, DockerOperationToolKit.ExecutePython)

    # Configure mock response
    mock_container = mock_docker_client.return_value.containers.run.return_value
    mock_container.exec_run.return_value = ExecResult(1, b"SyntaxError: invalid syntax")

    result = tool(code=code).call()  # pyright: ignore [reportCallIssue]
    assert "SyntaxError" in result


def test_execute_shell_with_error(
    docker_operation_toolkit: DockerOperationToolKit, mock_docker_client
):
    """Test executing invalid shell command."""
    _, tool = docker_operation_toolkit.create_tools()
    assert issubclass(tool, DockerOperationToolKit.ExecuteShell)

    # Configure mock response
    mock_container = mock_docker_client.return_value.containers.run.return_value
    mock_container.exec_run.return_value = ExecResult(127, b"command not found")

    result = tool(command="nonexistent_command").call()  # pyright: ignore [reportCallIssue]
    assert "not found" in result.lower()


def test_container_error_handling(
    docker_operation_toolkit: DockerOperationToolKit, mock_docker_client
):
    """Test error handling when container operations fail."""
    tool, _ = docker_operation_toolkit.create_tools()
    assert issubclass(tool, DockerOperationToolKit.ExecutePython)
    tool_instance = tool(code="print('test')")  # pyright: ignore [reportCallIssue]

    # Mock container to raise an exception
    mock_container = mock_docker_client.return_value.containers.run.return_value
    mock_container.exec_run.side_effect = APIError("Container stopped")

    result = tool_instance.call()
    assert "Error executing code:" in result


def test_execute_python_with_file_creation(
    docker_operation_toolkit: DockerOperationToolKit, mock_docker_client
):
    """Test executing Python code that creates files."""
    code = """
with open('test.txt', 'w') as f:
    f.write('Hello from Python!')
print('File created')
"""
    tool, _ = docker_operation_toolkit.create_tools()
    assert issubclass(tool, DockerOperationToolKit.ExecutePython)

    # Configure mock response
    mock_container = mock_docker_client.return_value.containers.run.return_value
    mock_container.exec_run.return_value = ExecResult(0, b"File created")

    result = tool(code=code).call()  # pyright: ignore [reportCallIssue]
    assert "File created" in result


def test_execute_shell_with_pipe(
    docker_operation_toolkit: DockerOperationToolKit, mock_docker_client
):
    """Test executing shell commands with pipe."""
    _, tool = docker_operation_toolkit.create_tools()
    assert issubclass(tool, DockerOperationToolKit.ExecuteShell)

    # Configure mock response
    mock_container = mock_docker_client.return_value.containers.run.return_value
    mock_container.exec_run.return_value = ExecResult(0, b"Hello")

    result = tool(command="echo 'Hello' | grep 'Hello'").call()  # pyright: ignore [reportCallIssue]
    assert "Hello" in result


def test_container_cleanup_on_error(mock_docker_client):
    """Test container cleanup when initialization fails."""
    # Configure mock to raise ImageNotFound
    mock_docker_client.return_value.containers.run.side_effect = ImageNotFound(
        "Image not found"
    )

    config = DockerOperationToolKitConfig(
        docker_image="nonexistent:latest",
        max_memory="256m",
        allow_network=False,
    )
    toolkit = DockerOperationToolKit(
        config=config, docker_container=DockerContainer(config=config)
    )
    tool, _ = toolkit.create_tools()
    assert issubclass(tool, DockerOperationToolKit.ExecutePython)
    result = tool(code="print('test')").call()  # pyright: ignore [reportCallIssue]
    assert "Error" in result


def test_usage_description(docker_operation_toolkit: DockerOperationToolKit):
    """Test usage description."""
    assert (
        docker_operation_toolkit.usage_description()
        == """- Tools for code execution:
    - ExecutePython: Executes Python code with optional requirements in a Docker container
    - ExecuteShell: Executes shell commands in a Docker container"""
    )



from mirascope.core.base.types import Usage


def get_common_usage(
    input_tokens: int | float | None,
    cached_tokens: int | float | None,
    output_tokens: int | float | None,
) -> Usage | None:
    """Get common usage from input and output tokens."""
    if input_tokens is None and cached_tokens is None and output_tokens is None:
        return None
    input_tokens = int(input_tokens or 0)
    cached_tokens = int(cached_tokens or 0)
    output_tokens = int(output_tokens or 0)
    return Usage(
        input_tokens=input_tokens,
        cached_tokens=cached_tokens,
        output_tokens=output_tokens,
        total_tokens=input_tokens + output_tokens,
    )



"""This module contains the `GoogleCallResponse` class.

usage docs: learn/calls.md#handling-responses
"""

from collections.abc import Sequence
from functools import cached_property

from google.genai.types import (
    ContentDict,
    ContentListUnion,
    ContentListUnionDict,
    FunctionResponseDict,
    GenerateContentResponse,
    GenerateContentResponseUsageMetadata,
    PartDict,
    # Import manually SchemaDict to avoid Pydantic error
    SchemaDict,  # noqa: F401
    Tool,
)
from pydantic import computed_field

from .. import BaseMessageParam
from ..base import BaseCallResponse, transform_tool_outputs
from ..base.types import CostMetadata, FinishReason, GoogleMetadata
from ._utils._convert_finish_reason_to_common_finish_reasons import (
    _convert_finish_reasons_to_common_finish_reasons,
)
from ._utils._message_param_converter import GoogleMessageParamConverter
from .call_params import GoogleCallParams
from .dynamic_config import GoogleDynamicConfig
from .tool import GoogleTool


class GoogleCallResponse(
    BaseCallResponse[
        GenerateContentResponse,
        GoogleTool,
        Tool,
        GoogleDynamicConfig,
        ContentListUnion | ContentListUnionDict,
        GoogleCallParams,
        ContentDict,
        GoogleMessageParamConverter,
    ]
):
    """A convenience wrapper around the Google API response.

    When calling the Google API using a function decorated with `google_call`, the
    response will be a `GoogleCallResponse` instance with properties that allow for
    more convenient access to commonly used attributes.


    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.google import google_call


    @google_call("google-1.5-flash")
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    response = recommend_book("fantasy")  # response is an `GoogleCallResponse` instance
    print(response.content)
    ```
    """

    _message_converter: type[GoogleMessageParamConverter] = GoogleMessageParamConverter

    _provider = "google"

    @computed_field
    @property
    def content(self) -> str:
        """Returns the contained string content for the 0th choice."""
        return self.response.candidates[0].content.parts[0].text  # pyright: ignore [reportOptionalSubscript, reportReturnType, reportOptionalMemberAccess, reportOptionalIterable]

    @computed_field
    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""

        return [
            candidate.finish_reason.value
            for candidate in (self.response.candidates or [])
            if candidate and candidate.finish_reason is not None
        ]

    @computed_field
    @property
    def model(self) -> str:
        """Returns the model name.

        google.generativeai does not return model, so we return the model provided by
        the user.
        """
        return (
            self.response.model_version if self.response.model_version else self._model
        )

    @computed_field
    @property
    def id(self) -> str | None:
        """Returns the id of the response.

        google.generativeai does not return an id
        """
        return None

    @property
    def usage(self) -> GenerateContentResponseUsageMetadata | None:
        """Returns the usage of the chat completion.

        google.generativeai does not have Usage, so we return None
        """
        return self.response.usage_metadata

    @computed_field
    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        return (
            self.response.usage_metadata.prompt_token_count
            if self.response.usage_metadata
            else None
        )

    @computed_field
    @property
    def cached_tokens(self) -> int | None:
        """Returns the number of cached tokens."""
        return (
            self.response.usage_metadata.cached_content_token_count
            if self.response.usage_metadata
            else None
        )

    @computed_field
    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        return (
            self.response.usage_metadata.candidates_token_count
            if self.response.usage_metadata
            else None
        )

    @computed_field
    @cached_property
    def message_param(self) -> ContentDict:
        """Returns the models's response as a message parameter."""
        return {"role": "model", "parts": self.response.candidates[0].content.parts}  # pyright: ignore [reportReturnType, reportOptionalSubscript, reportOptionalMemberAccess]

    @cached_property
    def tools(self) -> list[GoogleTool] | None:
        """Returns the list of tools for the response."""
        if self.tool_types is None:
            return None

        extracted_tools = []
        for tool_call in self.response.function_calls or []:
            for tool_type in self.tool_types:
                if tool_call.name == tool_type._name():  # pyright: ignore [reportOptionalMemberAccess]
                    extracted_tools.append(tool_type.from_tool_call(tool_call))  # pyright: ignore [reportArgumentType]
                    break

        return extracted_tools

    @cached_property
    def tool(self) -> GoogleTool | None:
        """Returns the 0th tool for the 0th candidate's 0th content part.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        tools = self.tools
        if tools:
            return tools[0]
        return None

    @classmethod
    @transform_tool_outputs
    def tool_message_params(
        cls, tools_and_outputs: Sequence[tuple[GoogleTool, str]]
    ) -> list[ContentDict]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The sequence of tools and their outputs from which the tool
                message parameters should be constructed.

        Returns:
            The list of constructed `FunctionResponse` parameters.
        """
        return [
            {
                "role": "user",
                "parts": [
                    PartDict(
                        function_response=FunctionResponseDict(
                            name=tool._name(), response={"result": output}
                        )
                    )
                    for tool, output in tools_and_outputs
                ],
            }
        ]

    @property
    def common_finish_reasons(self) -> list[FinishReason] | None:
        return _convert_finish_reasons_to_common_finish_reasons(self.finish_reasons)

    @property
    def common_message_param(self) -> BaseMessageParam:
        return GoogleMessageParamConverter.from_provider([self.message_param])[0]

    @property
    def common_user_message_param(self) -> BaseMessageParam | None:
        if not self.user_message_param:
            return None
        return GoogleMessageParamConverter.from_provider([self.user_message_param])[0]

    @property
    def cost_metadata(self) -> CostMetadata:
        cost_metadata = super().cost_metadata
        cost_metadata.google = GoogleMetadata(
            use_vertex_ai="use_vertex_ai" in self.metadata.get("tags", [])
        )
        return cost_metadata



"""The base type in a function as an LLM call to return for dynamic configuration."""

from collections.abc import Callable, Sequence
from typing import Any, Generic, TypeVar

from typing_extensions import NotRequired, TypedDict

from .call_params import BaseCallParams
from .metadata import Metadata
from .tool import BaseTool

_MessageParamT = TypeVar("_MessageParamT", bound=Any)
_CallParamsT = TypeVar("_CallParamsT", bound=BaseCallParams)
_ClientT = TypeVar("_ClientT", bound=object)


class DynamicConfigBase(TypedDict):
    metadata: NotRequired[Metadata]
    computed_fields: NotRequired[dict[str, Any | list[Any] | list[list[Any]]]]
    tools: NotRequired[Sequence[type[BaseTool] | Callable]]


class DynamicConfigMessages(DynamicConfigBase, Generic[_MessageParamT]):
    messages: NotRequired[Sequence[_MessageParamT]]


class DynamicConfigCallParams(DynamicConfigBase, Generic[_CallParamsT]):
    call_params: NotRequired[_CallParamsT]


class DynamicConfigClient(DynamicConfigBase, Generic[_ClientT]):
    client: NotRequired[_ClientT | None]


class DynamicConfigMessagesCallParams(
    DynamicConfigBase, Generic[_MessageParamT, _CallParamsT]
):
    messages: NotRequired[Sequence[_MessageParamT]]
    call_params: NotRequired[_CallParamsT]


class DynamicConfigMessagesClient(DynamicConfigBase, Generic[_MessageParamT, _ClientT]):
    messages: NotRequired[Sequence[_MessageParamT]]
    client: NotRequired[_ClientT | None]


class DynamicConfigCallParamsClient(DynamicConfigBase, Generic[_CallParamsT, _ClientT]):
    call_params: NotRequired[_CallParamsT]
    client: NotRequired[_ClientT | None]


class DynamicConfigFull(
    DynamicConfigBase, Generic[_MessageParamT, _CallParamsT, _ClientT]
):
    messages: NotRequired[Sequence[_MessageParamT]]
    call_params: NotRequired[_CallParamsT]
    client: NotRequired[_ClientT | None]


BaseDynamicConfig = (
    DynamicConfigBase
    | DynamicConfigMessages[_MessageParamT]
    | DynamicConfigCallParams[_CallParamsT]
    | DynamicConfigClient[_ClientT]
    | DynamicConfigMessagesCallParams[_MessageParamT, _CallParamsT]
    | DynamicConfigMessagesClient[_MessageParamT, _ClientT]
    | DynamicConfigCallParamsClient[_CallParamsT, _ClientT]
    | DynamicConfigFull[_MessageParamT, _CallParamsT, _ClientT]
    | None
)
"""The base type in a function as an LLM call to return for dynamic configuration.

Attributes:
    metadata: Any metadata to include in call responses.
    computed_fields: Fields to be computed and injected into the prompt template at
        runtime.
    tools: Tools to be provided to the LLM API call at runtime.
    messages: Custom message parameters, which will override any other form of writing
        prompts when used.
    call_params: Call parameters to use when making the LLM API call.
    client: A custom client to use in place of the default client.
"""



from functools import cached_property
from typing import Any
from unittest.mock import MagicMock, Mock, PropertyMock

import pytest
from pydantic import computed_field

from mirascope.core.base import (
    BaseCallParams,
    BaseCallResponse,
    BaseCallResponseChunk,
    BaseMessageParam,
    BaseStream,
    BaseTool,
)
from mirascope.core.base._utils import BaseMessageParamConverter
from mirascope.core.base.types import CostMetadata, FinishReason, Usage
from mirascope.llm.call_response import CallResponse
from mirascope.llm.call_response_chunk import CallResponseChunk
from mirascope.llm.stream import Stream
from mirascope.llm.tool import Tool


class DummyCallParams(BaseCallParams):
    pass


class DummyMessageParam(BaseMessageParam):
    role: str
    content: Any


class DummyTool(BaseTool):
    def call(self) -> str: ...

    @property
    def model_fields(self) -> list[str]:  # pyright: ignore [reportIncompatibleVariableOverride]
        # Return a list of field names actually used in your tool
        return ["field1"]

    field1: str = "value"


class DummyProviderResponse(
    BaseCallResponse[
        Any,
        DummyTool,
        Any,
        Any,
        DummyMessageParam,
        DummyCallParams,
        DummyMessageParam,
        Any,
    ]
):
    _message_converter: type = BaseMessageParamConverter

    @property
    def content(self) -> str: ...

    @property
    def finish_reasons(self) -> list[str] | None: ...

    @property
    def model(self) -> str | None: ...

    @property
    def id(self) -> str | None: ...

    @property
    def usage(self) -> Any: ...

    @property
    def input_tokens(self) -> int | float | None: ...

    @property
    def cached_tokens(self) -> int | float | None: ...

    @property
    def output_tokens(self) -> int | float | None: ...

    @property
    def cost(self) -> float | None: ...

    @computed_field
    @cached_property
    def message_param(self) -> DummyMessageParam: ...

    @cached_property
    def tools(self) -> list[DummyTool] | None: ...

    @cached_property
    def tool(self) -> DummyTool | None: ...

    @classmethod
    def tool_message_params(  # pyright: ignore [reportIncompatibleMethodOverride]
        cls, tools_and_outputs: list[tuple[DummyTool, str]]
    ) -> list[Any]: ...

    # common_ methods
    @property
    def common_finish_reasons(self) -> list[FinishReason] | None: ...

    @property
    def common_message_param(self) -> BaseMessageParam: ...

    @property
    def common_user_message_param(self) -> BaseMessageParam | None: ...

    @property
    def common_tools(self) -> list[Tool] | None: ...

    @property
    def common_usage(self) -> Usage: ...

    def common_construct_call_response(self) -> BaseCallResponse: ...
    def common_construct_message_param(
        self, tool_calls: list[Any] | None, content: str | None
    ) -> DummyMessageParam: ...

    @property
    def cost_metadata(self) -> CostMetadata: ...


class DummyProviderChunk(BaseCallResponseChunk[Any, FinishReason]):
    @property
    def content(self) -> str: ...

    @property
    def finish_reasons(self) -> list[FinishReason] | None: ...

    @property
    def model(self) -> str | None: ...

    @property
    def id(self) -> str | None: ...

    @property
    def usage(self) -> Any: ...

    @property
    def input_tokens(self) -> int | float | None: ...

    @property
    def cached_tokens(self) -> int | float | None: ...

    @property
    def output_tokens(self) -> int | float | None: ...

    @property
    def common_finish_reasons(self) -> list[FinishReason] | None: ...

    @property
    def common_usage(self) -> Usage: ...

    @property
    def cost_metadata(self) -> CostMetadata: ...


class DummyStream(Stream): ...


@pytest.mark.asyncio
async def test_stream():
    def sync_gen():
        mock_chunk = DummyProviderChunk(chunk=Mock())
        type(mock_chunk).content = PropertyMock(return_value="fake chunk content")  # pyright: ignore [reportAttributeAccessIssue]
        yield (mock_chunk, DummyTool())

    async def async_gen(self):
        mock_chunk = DummyProviderChunk(chunk=Mock())
        type(mock_chunk).content = PropertyMock(return_value="fake chunk content")  # pyright: ignore [reportAttributeAccessIssue]
        yield (mock_chunk, DummyTool())

    mock_stream = MagicMock(spec=BaseStream)
    mock_stream.model = "test_model"
    mock_stream.cost = 0.02
    mock_stream.cost_metadata = CostMetadata(streaming_mode=True)

    mock_stream.construct_call_response.return_value = DummyProviderResponse(
        metadata={},
        response={},
        prompt_template="",
        fn_args={},
        dynamic_config={},
        messages=[],
        call_params={},
        call_kwargs={},
        start_time=0,
        end_time=0,
    )

    dummy_stream_instance = DummyStream(stream=mock_stream)

    mock_stream.__iter__.return_value = sync_gen()
    for chunk, tool in dummy_stream_instance:
        assert isinstance(chunk, CallResponseChunk)
        assert isinstance(tool, Tool)

    mock_stream.__aiter__ = async_gen
    # dummy_stream_instance.stream = async_gen()
    async for chunk, tool in dummy_stream_instance:
        assert isinstance(chunk, CallResponseChunk)
        assert isinstance(tool, Tool)

    assert dummy_stream_instance.model == "test_model"
    assert dummy_stream_instance.cost == 0.02
    assert dummy_stream_instance.cost_metadata == CostMetadata(streaming_mode=True)

    call_response_instance = dummy_stream_instance.construct_call_response()
    assert isinstance(call_response_instance, CallResponse)


@pytest.fixture
def dummy_stream_instance() -> DummyStream:
    def sync_gen(): ...

    mock_stream = MagicMock()
    mock_stream.cost = 0.02

    mock_stream.construct_call_response.return_value = DummyProviderResponse(
        metadata={},
        response={},
        prompt_template="",
        fn_args={},
        dynamic_config={},
        messages=[],
        call_params={},
        call_kwargs={},
        start_time=0,
        end_time=0,
    )
    mock_stream._construct_message_param.return_value = DummyMessageParam(
        role="assistant", content="test content"
    )

    mock_stream.__iter__ = sync_gen()
    dummy_stream = DummyStream(stream=mock_stream)

    return dummy_stream


def test_construct_message_param(dummy_stream_instance: DummyStream):
    tool_calls = [MagicMock()]
    content = "test content"
    message_param = dummy_stream_instance._construct_message_param(tool_calls, content)  # pyright: ignore [reportArgumentType]
    assert isinstance(message_param, DummyMessageParam), (
        "Should return a DummyMessageParam instance"
    )


def test_construct_call_response(dummy_stream_instance: DummyStream):
    call_resp = dummy_stream_instance.construct_call_response()  # pyright: ignore [reportAbstractUsage]
    assert isinstance(call_resp, CallResponse), "Should return a CallResponse instance"


def test_tool_message_params(dummy_stream_instance: DummyStream):
    tools_and_outputs = [(DummyTool(), "test output")]
    result = dummy_stream_instance.tool_message_params(tools_and_outputs)  # pyright: ignore [reportArgumentType]

    assert isinstance(result, list), "tool_message_params should return a list"



"""Tests the `google._utils.handle_stream` module."""

import pytest
from google.genai.types import (
    Candidate,
    Content,
    FunctionCall,
    GenerateContentResponse,
    Part,
)
from google.genai.types import (
    FinishReason as GoogleFinishReason,
)

from mirascope.core.google._utils._handle_stream import (
    handle_stream,
    handle_stream_async,
)
from mirascope.core.google.tool import GoogleTool


class FormatBook(GoogleTool):
    """Returns the title and author nicely formatted."""

    title: str
    author: str

    def call(self) -> None:
        """Dummy call."""


@pytest.fixture()
def mock_chunks() -> list[GenerateContentResponse]:
    """Returns a list of mock `GenerateContentResponse` instances."""
    return [
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=GoogleFinishReason.STOP,
                    content=Content(
                        parts=[Part(text="The author is ")],
                        role="model",
                    ),
                )
            ]
        ),
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=GoogleFinishReason.STOP,
                    content=Content(
                        parts=[Part(text="Patrick Rothfuss")],
                        role="model",
                    ),
                )
            ]
        ),
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=GoogleFinishReason.STOP,
                    content=Content(
                        parts=[
                            Part(
                                function_call=FunctionCall(
                                    name="FormatBook",
                                    args={
                                        "title": "The Name of the Wind",
                                        "author": "Patrick Rothfuss",
                                    },
                                )
                            )
                        ],
                        role="model",
                    ),
                )
            ]
        ),
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=GoogleFinishReason.STOP,
                    content=Content(
                        parts=[
                            Part(
                                function_call=FunctionCall(
                                    name="FormatBook",
                                    args={
                                        "title": "The Name of the Wind",
                                        "author": "Patrick Rothfuss",
                                    },
                                )
                            ),
                            Part(
                                function_call=FunctionCall(
                                    name="FormatBook",
                                    args={
                                        "title": "Mistborn: The Final Empire",
                                        "author": "Brandon Sanderson",
                                    },
                                )
                            ),
                        ],
                        role="model",
                    ),
                )
            ]
        ),
    ]


def test_handle_stream(mock_chunks: list[GenerateContentResponse]) -> None:
    """Tests the `handle_stream` function."""
    result = list(
        handle_stream((chunk for chunk in mock_chunks), tool_types=[FormatBook])
    )

    assert len(result) == 5
    assert result[0][0].content == "The author is "
    assert result[0][1] is None
    assert result[1][0].content == "Patrick Rothfuss"
    assert result[1][1] is None
    assert (tool := result[2][1]) is not None
    assert isinstance(tool, FormatBook)
    assert tool.model_dump(exclude={"tool_call", "delta"}) == {
        "author": "Patrick Rothfuss",
        "title": "The Name of the Wind",
    }
    assert (tool := result[3][1]) is not None
    assert isinstance(tool, FormatBook)
    assert tool.model_dump(exclude={"tool_call", "delta"}) == {
        "author": "Patrick Rothfuss",
        "title": "The Name of the Wind",
    }
    assert (tool := result[4][1]) is not None
    assert isinstance(tool, FormatBook)
    assert tool.model_dump(exclude={"tool_call", "delta"}) == {
        "author": "Brandon Sanderson",
        "title": "Mistborn: The Final Empire",
    }


@pytest.mark.asyncio
async def test_handle_stream_async(
    mock_chunks: list[GenerateContentResponse],
) -> None:
    """Tests the `handle_stream_async` function."""
    result = []

    async def generator():
        for chunk in mock_chunks:
            yield chunk

    async for chunk, tool in handle_stream_async(generator(), tool_types=[FormatBook]):
        result.append((chunk, tool))

    assert len(result) == 5
    assert result[0][0].content == "The author is "
    assert result[0][1] is None
    assert result[1][0].content == "Patrick Rothfuss"
    assert result[1][1] is None
    assert (tool := result[2][1]) is not None
    assert isinstance(tool, FormatBook)
    assert tool.model_dump(exclude={"tool_call", "delta"}) == {
        "author": "Patrick Rothfuss",
        "title": "The Name of the Wind",
    }
    assert (tool := result[3][1]) is not None
    assert isinstance(tool, FormatBook)
    assert tool.model_dump(exclude={"tool_call", "delta"}) == {
        "author": "Patrick Rothfuss",
        "title": "The Name of the Wind",
    }
    assert (tool := result[4][1]) is not None
    assert isinstance(tool, FormatBook)
    assert tool.model_dump(exclude={"tool_call", "delta"}) == {
        "author": "Brandon Sanderson",
        "title": "Mistborn: The Final Empire",
    }



"""Tests the `cohere._utils.calculate_cost` function."""

from mirascope.core.base.types import CostMetadata
from mirascope.core.costs._cohere_calculate_cost import calculate_cost


def test_calculate_cost() -> None:
    """Tests the `calculate_cost` function."""
    assert calculate_cost(CostMetadata(), model="command-r-plus") is None
    assert (
        calculate_cost(CostMetadata(input_tokens=1, output_tokens=1), model="unknown")
        is None
    )
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1, output_tokens=1), model="command-r-plus"
        )
        == 1.8e-05
    )



from typing import cast

from cohere.types import ChatMessage

from mirascope.core import BaseMessageParam
from mirascope.core.base import TextPart, ToolCallPart
from mirascope.core.base._utils._base_message_param_converter import (
    BaseMessageParamConverter,
)
from mirascope.core.cohere._utils import convert_message_params


class CohereMessageParamConverter(BaseMessageParamConverter):
    """Converts between Cohere `ChatMessage` and Mirascope `BaseMessageParam`."""

    @staticmethod
    def to_provider(message_params: list[BaseMessageParam]) -> list[ChatMessage]:
        """
        Convert from Mirascope `BaseMessageParam` to Cohere's `ChatMessage`.
        """
        return convert_message_params(
            cast(list[BaseMessageParam | ChatMessage], message_params)
        )

    @staticmethod
    def from_provider(message_params: list[ChatMessage]) -> list[BaseMessageParam]:
        """
        Convert from Cohere's `ChatMessage` to Mirascope `BaseMessageParam`.
        """
        converted = []
        for message_param in message_params:
            role = getattr(message_param, "role", "assistant")
            if not message_param.tool_calls:
                converted.append(
                    BaseMessageParam(role=role, content=message_param.message)
                )
                continue

            converted_content = []

            if message_param.message:
                converted_content.append(
                    TextPart(type="text", text=message_param.message)
                )

            for tool_call in message_param.tool_calls:
                converted_content.append(
                    ToolCallPart(
                        type="tool_call", name=tool_call.name, args=tool_call.parameters
                    )
                )

            converted.append(BaseMessageParam(role=role, content=converted_content))
        return converted



from unittest.mock import AsyncMock, MagicMock, patch

import aiobotocore.client
import pytest
from aiobotocore.session import AioSession
from pydantic import BaseModel

from mirascope.core.bedrock._utils._convert_common_call_params import (
    convert_common_call_params,
)
from mirascope.core.bedrock._utils._setup_call import (
    _AsyncBedrockRuntimeWrappedClient,
    _extract_async_stream_fn,
    _extract_sync_stream_fn,
    setup_call,
)
from mirascope.core.bedrock.tool import BedrockTool


@pytest.fixture()
def mock_base_setup_call() -> MagicMock:
    mock_setup_call = MagicMock()
    mock_setup_call.return_value = [MagicMock() for _ in range(3)] + [{}]
    return mock_setup_call


def test_extract_sync_stream_fn():
    mock_fn = MagicMock()
    mock_fn.return_value = {
        "ResponseMetadata": {"RequestId": "123"},
        "stream": [
            {
                "contentBlockDelta": {"delta": [{"text": "Hello"}]},
            },
            {
                "contentBlockDelta": {"delta": [{"text": "World"}]},
            },
        ],
    }

    extracted_fn = _extract_sync_stream_fn(mock_fn, "test-model")
    result = list(extracted_fn())

    assert len(result) == 2
    assert result[0]["contentBlockDelta"]["delta"][0]["text"] == "Hello"  # pyright: ignore [reportTypedDictNotRequiredAccess, reportGeneralTypeIssues]
    assert result[1]["contentBlockDelta"]["delta"][0]["text"] == "World"  # pyright: ignore [reportTypedDictNotRequiredAccess, reportGeneralTypeIssues]
    assert all(chunk["responseMetadata"] == {"RequestId": "123"} for chunk in result)
    assert all(chunk["model"] == "test-model" for chunk in result)


@pytest.mark.asyncio
async def test_extract_async_stream_fn():
    async def async_generator():
        yield {"chunk1": "async_data1"}
        yield {"chunk2": "async_data2"}

    mock_fn = AsyncMock()
    mock_fn.return_value = {
        "ResponseMetadata": {"RequestId": "456"},
        "stream": async_generator(),
    }

    extracted_fn = _extract_async_stream_fn(mock_fn, "test-async-model")
    result = [chunk async for chunk in extracted_fn()]

    assert len(result) == 2
    assert result[0]["chunk1"] == "async_data1"  # pyright: ignore [reportGeneralTypeIssues]
    assert result[1]["chunk2"] == "async_data2"  # pyright: ignore [reportGeneralTypeIssues]
    assert all(chunk["responseMetadata"] == {"RequestId": "456"} for chunk in result)
    assert all(chunk["model"] == "test-async-model" for chunk in result)


@patch(
    "mirascope.core.bedrock._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value[3] = {}
    fn = MagicMock()
    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert prompt_template == mock_base_setup_call.return_value[0]
    assert tool_types == mock_base_setup_call.return_value[2]
    assert "modelId" in call_kwargs and call_kwargs["modelId"] == "anthropic.claude-v2"
    assert "messages" in call_kwargs and call_kwargs["messages"] == messages
    mock_base_setup_call.assert_called_once_with(
        fn, {}, None, None, BedrockTool, {}, convert_common_call_params
    )
    mock_convert_message_params.assert_called_once_with(
        mock_base_setup_call.return_value[1]
    )
    assert messages == mock_convert_message_params.return_value


@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_system_message(
    mock_utils: MagicMock, mock_base_setup_call: MagicMock
) -> None:
    mock_base_setup_call.return_value[1] = [
        {"role": "system", "content": [{"text": "system test"}]},
        {"role": "user", "content": [{"text": "user test"}]},
    ]
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value[3] = {}
    _, _, messages, _, _ = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert messages[0]["role"] == "user"
    assert messages[0]["content"] == [{"text": "user test"}]


@patch(
    "mirascope.core.bedrock._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_json_mode(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    mock_utils.setup_call = mock_base_setup_call
    mock_utils.json_mode_content = MagicMock(return_value="json_content")
    mock_base_setup_call.return_value[1] = [
        {"role": "user", "content": [{"text": "test"}]}
    ]
    mock_base_setup_call.return_value[3] = {"tools": MagicMock()}
    mock_convert_message_params.side_effect = lambda x: x
    _, _, messages, _, call_kwargs = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert messages[-1]["content"] == [{"text": "testjson_content"}]
    assert "tools" not in call_kwargs


@patch(
    "mirascope.core.bedrock._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_json_mode_no_text(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    mock_utils.setup_call = mock_base_setup_call
    mock_utils.json_mode_content = MagicMock(return_value="json_content")
    mock_base_setup_call.return_value[1] = [
        {"role": "user", "content": [{"image": "test_image"}]}
    ]
    mock_base_setup_call.return_value[3] = {"tools": MagicMock()}
    mock_convert_message_params.side_effect = lambda x: x
    _, _, messages, _, call_kwargs = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert messages[-1]["content"] == [
        {"image": "test_image"},
        {"text": "json_content"},
    ]
    assert "tools" not in call_kwargs


@patch(
    "mirascope.core.bedrock._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_extract(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    tool_mock = MagicMock(spec=BedrockTool, __name__="tool")
    tool_mock._name.return_value = "mock_tool"
    mock_base_setup_call.return_value[2] = [tool_mock]
    mock_base_setup_call.return_value[3] = {"toolConfig": {}}
    mock_utils.setup_call = mock_base_setup_call
    _, _, _, tool_types, call_kwargs = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=BaseModel,
        stream=False,
    )
    assert isinstance(tool_types, list)
    assert "toolConfig" in call_kwargs and call_kwargs["toolConfig"] == {
        "toolChoice": {
            "tool": {
                "name": "mock_tool",
            }
        }
    }


@patch(
    "mirascope.core.bedrock._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_with_tools(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value[3] = {"tools": [{"name": "test_tool"}]}
    fn = MagicMock()
    _, _, _, _, call_kwargs = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert "toolConfig" in call_kwargs
    assert call_kwargs["toolConfig"] == {"tools": [{"name": "test_tool"}]}


@patch("mirascope.core.bedrock._utils._setup_call.get_session")
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_client_creation(
    mock_utils: MagicMock,
    mock_get_session: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value[1] = [
        {"role": "user", "content": [{"text": "user test"}]},
    ]
    mock_base_setup_call.return_value[3] = {}

    # Test sync client creation
    setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )

    # Test async client creation
    async def async_fn(): ...

    setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=async_fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )

    mock_get_session.assert_called_once()

    # Test when client is provided
    mock_client = MagicMock()
    setup_call(
        model="anthropic.claude-v2",
        client=mock_client,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )


@patch("mirascope.core.bedrock._utils._setup_call.get_async_create_fn")
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_aiobotocore_client(
    mock_utils: MagicMock,
    mock_get_async_create_fn: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """
    Test that when the provided client is an instance of aiobotocore.client.AioBaseClient,
    setup_call uses get_async_create_fn with client.converse and the _extract_async_stream_fn.
    """
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value = [
        "prompt",
        [{"role": "user", "content": [{"text": "test"}]}],
        None,
        {},
    ]

    mock_aiobotocore_client = MagicMock(spec=aiobotocore.client.AioBaseClient)
    mock_aiobotocore_client.converse = AsyncMock()
    mock_aiobotocore_client.converse_stream = AsyncMock()

    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="anthropic.claude-v2",
        client=mock_aiobotocore_client,
        fn=MagicMock(),  # can be any callable
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )

    mock_get_async_create_fn.assert_called_once()
    args, kwargs = mock_get_async_create_fn.call_args

    assert args[0] is mock_aiobotocore_client.converse
    assert callable(args[1])

    assert create == mock_get_async_create_fn.return_value

    assert prompt_template == "prompt"
    assert messages[0]["role"] == "user"
    assert call_kwargs["modelId"] == "anthropic.claude-v2"


@patch("mirascope.core.bedrock._utils._setup_call.get_async_create_fn")
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_async_bedrock_runtime_wrapped_client(
    mock_utils: MagicMock,
    mock_get_async_create_fn: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """
    Test that when the provided client is an instance of _AsyncBedrockRuntimeWrappedClient,
    setup_call uses get_async_create_fn correctly with client.converse and client.converse_stream.
    """
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value = [
        "prompt",
        [{"role": "user", "content": [{"text": "test"}]}],
        None,
        {},
    ]

    wrapped_client = _AsyncBedrockRuntimeWrappedClient(
        MagicMock(), "anthropic.claude-v2"
    )

    create, prompt_template, messages, tool_types, call_kwargs = setup_call(  # pyright: ignore [reportCallIssue]
        model="anthropic.claude-v2",
        client=wrapped_client,  # pyright: ignore [reportArgumentType]
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )

    mock_get_async_create_fn.assert_called_once_with(
        wrapped_client.converse,
        wrapped_client.converse_stream,
    )
    assert create == mock_get_async_create_fn.return_value
    assert prompt_template == "prompt"
    assert messages[0]["role"] == "user"
    assert "modelId" in call_kwargs and call_kwargs["modelId"] == "anthropic.claude-v2"


@pytest.mark.asyncio
async def test_async_bedrock_runtime_wrapped_client_converse():
    """
    Test that _AsyncBedrockRuntimeWrappedClient.converse calls client.converse inside the context
    manager and returns the response.
    """
    mock_session = MagicMock(spec=AioSession)
    mock_client = AsyncMock()
    mock_client.converse = AsyncMock(return_value={"some": "result"})

    mock_context_manager = AsyncMock()
    mock_context_manager.__aenter__.return_value = mock_client
    mock_context_manager.__aexit__.return_value = None

    mock_session.create_client.return_value = mock_context_manager

    wrapped_client = _AsyncBedrockRuntimeWrappedClient(mock_session, "test-model")
    result = await wrapped_client.converse(param1="value1")  # pyright: ignore [reportCallIssue]

    assert result == {"some": "result"}

    mock_client.converse.assert_called_once_with(param1="value1")

    mock_session.create_client.assert_called_once_with("bedrock-runtime")


@pytest.mark.asyncio
async def test_async_bedrock_runtime_wrapped_client_converse_stream():
    """
    Test that _AsyncBedrockRuntimeWrappedClient.converse_stream calls client.converse_stream
    and yields dicts with the correct keys and values, aligning with the AsyncStreamOutputChunk structure.
    """
    mock_session = MagicMock(spec=AioSession)
    mock_client = AsyncMock()

    async def fake_stream():
        yield {"content": "chunk1"}
        yield {"content": "chunk2"}

    mock_client.converse_stream = AsyncMock(
        return_value={
            "ResponseMetadata": {"request": "id"},
            "stream": fake_stream(),
        }
    )

    mock_context_manager = AsyncMock()
    mock_context_manager.__aenter__.return_value = mock_client
    mock_context_manager.__aexit__.return_value = None
    mock_session.create_client.return_value = mock_context_manager

    wrapped_client = _AsyncBedrockRuntimeWrappedClient(mock_session, "test-model")

    chunks = []
    async for chunk in wrapped_client.converse_stream(param2="value2"):  # pyright: ignore [reportCallIssue]
        chunks.append(chunk)

    assert len(chunks) == 2

    for i, expected_content in enumerate(["chunk1", "chunk2"]):
        assert isinstance(chunks[i], dict)
        assert chunks[i]["model"] == "test-model"
        assert chunks[i]["content"] == expected_content
        assert chunks[i]["responseMetadata"] == {"request": "id"}

    mock_client.converse_stream.assert_called_once_with(param2="value2")

    mock_session.create_client.assert_called_once_with("bedrock-runtime")


@patch("mirascope.core.bedrock._utils._setup_call.Session")
@patch("mirascope.core.bedrock._utils._setup_call.get_session")
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_env_vars(
    mock_utils: MagicMock,
    mock_get_session: MagicMock,
    mock_session_class: MagicMock,
    mock_base_setup_call: MagicMock,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test that environment variables are properly passed to Session and get_session."""
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value[1] = [
        {"role": "user", "content": [{"text": "user test"}]},
    ]
    mock_base_setup_call.return_value[3] = {}

    # Set environment variables
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "test_access_key")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "test_secret_key")
    monkeypatch.setenv("AWS_SESSION_TOKEN", "test_session_token")
    monkeypatch.setenv("AWS_REGION_NAME", "us-west-2")
    monkeypatch.setenv("AWS_PROFILE", "test_profile")

    # Test sync client creation with env vars
    setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )

    # Verify Session was called with expected env vars
    mock_session_class.assert_called_once_with(
        aws_access_key_id="test_access_key",
        aws_secret_access_key="test_secret_key",
        aws_session_token="test_session_token",
        region_name="us-west-2",
        profile_name="test_profile",
    )

    # Test async client creation with env vars
    async def async_fn(): ...

    mock_session_class.reset_mock()

    setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=async_fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )

    # Verify get_session was called with expected env vars
    mock_get_session.assert_called_with(
        env_vars={
            "aws_access_key_id": "test_access_key",
            "aws_secret_access_key": "test_secret_key",
            "aws_session_token": "test_session_token",
            "region_name": "us-west-2",
            "profile_name": "test_profile",
        }
    )



"""This module contains the `GoogleCallResponseChunk` class.

usage docs: learn/streams.md#handling-streamed-responses
"""

from typing import cast

from google.genai.types import FinishReason as GoogleFinishReason
from google.genai.types import (
    GenerateContentResponse,
    GenerateContentResponseUsageMetadata,
)

from mirascope.core.base.types import CostMetadata, FinishReason

from ..base import BaseCallResponseChunk
from ._utils._convert_finish_reason_to_common_finish_reasons import (
    _convert_finish_reasons_to_common_finish_reasons,
)


class GoogleCallResponseChunk(
    BaseCallResponseChunk[GenerateContentResponse, GoogleFinishReason]
):
    """A convenience wrapper around the Google API streamed response chunks.

    When calling the Google API using a function decorated with `google_call` and
    `stream` set to `True`, the stream will contain `GoogleCallResponseChunk` instances

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.google import google_call


    @google_call("google-1.5-flash", stream=True)
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    stream = recommend_book("fantasy")  # response is an `GoogleStream`
    for chunk, _ in stream:
        print(chunk.content, end="", flush=True)
    ```
    """

    @property
    def content(self) -> str:
        """Returns the chunk content for the 0th choice."""
        if (
            not (candidates := self.chunk.candidates)
            or not (content := candidates[0].content)
            or not (parts := content.parts)
            or not (text := parts[0].text)
        ):
            return ""
        return text

    @property
    def finish_reasons(self) -> list[GoogleFinishReason]:
        """Returns the finish reasons of the response."""
        return [
            candidate.finish_reason
            for candidate in (self.chunk.candidates or [])
            if candidate.finish_reason is not None
        ]

    @property
    def model(self) -> str | None:
        """Returns the model name.

        google.generativeai does not return model, so we return None
        """
        return self.chunk.model_version

    @property
    def id(self) -> str | None:
        """Returns the id of the response.

        google.generativeai does not return an id
        """
        return None

    @property
    def usage(self) -> GenerateContentResponseUsageMetadata | None:
        """Returns the usage of the chat completion.

        google.generativeai does not have Usage, so we return None
        """
        return self.chunk.usage_metadata

    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        return self.usage.prompt_token_count if self.usage else None

    @property
    def cached_tokens(self) -> int | None:
        """Returns the number of cached tokens."""
        return self.usage.cached_content_token_count if self.usage else None

    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        return self.usage.candidates_token_count if self.usage else None

    @property
    def cost_metadata(self) -> CostMetadata:
        """Returns the cost metadata."""
        return super().cost_metadata

    @property
    def common_finish_reasons(self) -> list[FinishReason] | None:
        return _convert_finish_reasons_to_common_finish_reasons(
            cast(list[str], self.finish_reasons)
        )



"""Tests for the _context.py module."""

from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseModel

from mirascope.core.base import BaseTool, CommonCallParams
from mirascope.llm._context import (
    CallArgs,
    LLMContext,
    _context,
    apply_context_overrides_to_call_args,
    context,
    get_current_context,
)


class MockTool(BaseTool):
    """Mock tool for testing."""

    name: str = "mock_tool"
    description: str = "Mock tool for testing"


class MockResponseModel(BaseModel):
    """Mock response model for testing."""

    value: str


def test_basic_context_enter_exit():
    """Test basic enter and exit functionality of LLMContext."""
    # Test enter
    ctx = LLMContext(provider="openai", model="gpt-4")
    with ctx as entered_ctx:
        assert entered_ctx is ctx
        assert get_current_context() is ctx

    # Test exit - context should be cleared
    assert get_current_context() is None


def test_context_no_old_context():
    """Test creating a context when there's no old context."""
    with _context(
        provider="openai",
        model="gpt-4",
        stream=True,
        tools=[MockTool],
        response_model=MockResponseModel,
        output_parser=lambda x: x,
        json_mode=True,
        client=MagicMock(),
        call_params=CommonCallParams(),
    ) as ctx:
        assert ctx.provider == "openai"
        assert ctx.model == "gpt-4"
        assert ctx.stream is True
        assert ctx.tools == [MockTool]
        assert ctx.response_model == MockResponseModel
        assert ctx.output_parser is not None
        assert ctx.json_mode is True
        assert ctx.client is not None
        assert ctx.call_params is not None


def test_context_with_old_context():
    """Test creating a nested context with an existing context."""
    with (
        _context(
            provider="openai",
            model="gpt-3.5-turbo",
            stream=False,
            tools=None,
            response_model=None,
            output_parser=None,
            json_mode=False,
            client=None,
            call_params=None,
        ),
        _context(
            provider="anthropic",
            model="claude-3-opus",
            stream=True,
            tools=None,
            response_model=None,
            output_parser=None,
            json_mode=None,
            client=None,
            call_params=None,
        ) as ctx2,
    ):
        # Now create a new context inside the first one
        # The new context should override fields that were set
        assert ctx2.provider == "anthropic"
        assert ctx2.model == "claude-3-opus"
        assert ctx2.stream is True

        # Other fields should have remained None since they were not set in either context
        assert ctx2.tools is None
        assert ctx2.response_model is None


def test_nested_context_with_partial_override():
    """Test creating a nested context with only some fields overridden."""
    mock_client1 = MagicMock()
    mock_client2 = MagicMock()
    call_params1 = CommonCallParams()

    with (
        _context(
            provider="openai",
            model="gpt-4o",
            stream=False,
            tools=[MockTool],
            response_model=MockResponseModel,
            output_parser=lambda x: x,
            json_mode=True,
            client=mock_client1,
            call_params=call_params1,
        ) as ctx1,
        _context(
            provider="openai",  # same provider
            model="gpt-4o-mini",  # different model
            stream=None,  # not overridden
            tools=None,  # not overridden
            response_model=None,  # not overridden
            output_parser=None,  # not overridden
            json_mode=None,  # not overridden
            client=mock_client2,  # different client
            call_params=None,  # not overridden
        ) as ctx2,
    ):
        # Check overridden fields
        assert ctx2.provider == "openai"
        assert ctx2.model == "gpt-4o-mini"
        assert ctx2.client is mock_client2

        # Check inherited fields
        assert ctx2.stream is ctx1.stream
        assert ctx2.tools is ctx1.tools
        assert ctx2.response_model is ctx1.response_model
        assert ctx2.output_parser is ctx1.output_parser
        assert ctx2.json_mode is ctx1.json_mode
        assert ctx2.call_params is ctx1.call_params


def test_apply_context_overrides_to_call_args_no_context():
    """Test applying context overrides when no context is active."""
    call_args: CallArgs = {
        "provider": "openai",
        "model": "gpt-4",
        "stream": False,
        "tools": None,
        "response_model": None,
        "output_parser": None,
        "json_mode": False,
        "client": None,
        "call_params": None,
    }

    # With no active context, call_args should be returned unchanged
    with patch("mirascope.llm._context.get_current_context", return_value=None):
        result = apply_context_overrides_to_call_args(call_args)
        assert result == call_args


def test_apply_context_overrides_to_call_args_structural_overrides():
    """Test applying context with structural overrides."""
    # Original call args
    call_args: CallArgs = {
        "provider": "openai",
        "model": "gpt-4",
        "stream": False,
        "tools": [MockTool],
        "response_model": None,
        "output_parser": None,
        "json_mode": False,
        "client": None,
        "call_params": None,
    }

    # Create a context with structural overrides
    with patch("mirascope.llm._context.get_current_context") as mock_get_context:
        # Mock a context with stream override (structural)
        mock_get_context.return_value = LLMContext(
            provider="anthropic",
            model="claude-3",
            stream=True,  # structural override
            response_model=MockResponseModel,  # structural override
            output_parser=lambda x: x,  # structural override
        )

        result = apply_context_overrides_to_call_args(call_args)

        # Check that overrides were applied
        assert result["provider"] == "anthropic"
        assert result["model"] == "claude-3"
        assert result["stream"] is True
        assert result["response_model"] == MockResponseModel
        assert result["output_parser"] is not None

        # Check that we reset other structural fields
        assert result["tools"] is None


def test_apply_context_overrides_to_call_args_response_model_override():
    """Test applying context with response_model override specifically."""
    # Original call args
    call_args: CallArgs = {
        "provider": "openai",
        "model": "gpt-4",
        "stream": False,
        "tools": [MockTool],
        "response_model": None,
        "output_parser": None,
        "json_mode": False,
        "client": None,
        "call_params": None,
    }

    # Create a context with response_model override
    with patch("mirascope.llm._context.get_current_context") as mock_get_context:
        # Mock a context with response_model override (forces tools to None)
        mock_get_context.return_value = LLMContext(
            provider="openai",
            model="gpt-4",
            response_model=MockResponseModel,  # This should reset tools
        )

        result = apply_context_overrides_to_call_args(call_args)

        # Check that response_model override forces tools to None
        assert result["response_model"] == MockResponseModel
        assert result["tools"] is None


def test_apply_context_overrides_to_call_args_all_possible_overrides():
    """Test applying context with all possible overrides."""
    # Original call args
    call_args: CallArgs = {
        "provider": "openai",
        "model": "gpt-4",
        "stream": False,
        "tools": None,
        "response_model": None,
        "output_parser": None,
        "json_mode": False,
        "client": None,
        "call_params": None,
    }

    # Create a mock for client and call params
    mock_client = MagicMock()
    call_params = CommonCallParams(temperature=0.8)

    # Create a context with all possible overrides
    with patch("mirascope.llm._context.get_current_context") as mock_get_context:
        # Mock a context with all fields overridden
        mock_context = LLMContext(
            provider="anthropic",
            model="claude-3",
            stream=True,
            tools=[MockTool],
            response_model=MockResponseModel,
            output_parser=lambda x: x,
            json_mode=True,
            client=mock_client,
            call_params=call_params,
        )
        mock_get_context.return_value = mock_context

        result = apply_context_overrides_to_call_args(call_args)

        # Check that all overrides were applied
        assert result["provider"] == "anthropic"
        assert result["model"] == "claude-3"
        assert result["stream"] is True
        assert result["tools"] is mock_context.tools
        assert result["response_model"] is mock_context.response_model
        assert result["output_parser"] is mock_context.output_parser
        assert result["json_mode"] is mock_context.json_mode
        assert result["client"] is mock_context.client
        assert result["call_params"] is mock_context.call_params


def test_context_function_validation():
    """Test the context function validation."""
    # Test that it raises error if only provider is provided
    with pytest.raises(ValueError):
        context(provider="openai", model=None)  # pyright: ignore [reportCallIssue,reportArgumentType]

    # Test that it raises error if only model is provided
    with pytest.raises(ValueError):
        context(provider=None, model="gpt-4")  # pyright: ignore [reportCallIssue,reportArgumentType]


def test_context_function():
    """Test the public context function."""
    with patch("mirascope.llm._context._context") as mock_context:
        mock_context.return_value = LLMContext(provider="openai", model="gpt-4")

        with context(provider="openai", model="gpt-4"):
            pass

        # Verify that _context was called with the right parameters
        mock_context.assert_called_once_with(
            provider="openai", model="gpt-4", client=None, call_params=None
        )


def test_context_function_with_client_and_params():
    """Test the public context function with client and call parameters."""
    mock_client = MagicMock()
    call_params = CommonCallParams(temperature=0.5)

    with patch("mirascope.llm._context._context") as mock_context:
        mock_context.return_value = LLMContext(
            provider="openai",
            model="gpt-4",
            client=mock_client,
            call_params=call_params,
        )

        with context(
            provider="openai",
            model="gpt-4",
            client=mock_client,
            call_params=call_params,
        ):
            pass

        # Verify that _context was called with the right parameters
        mock_context.assert_called_once_with(
            provider="openai",
            model="gpt-4",
            client=mock_client,
            call_params=call_params,
        )


def test_apply_context_overrides_to_call_args_with_explicit_context():
    """Test applying context overrides with an explicitly provided context."""
    # Original call args
    call_args: CallArgs = {
        "provider": "openai",
        "model": "gpt-4",
        "stream": False,
        "tools": None,
        "response_model": None,
        "output_parser": None,
        "json_mode": False,
        "client": None,
        "call_params": None,
    }

    # Create an explicit context
    explicit_context = LLMContext(
        provider="anthropic",
        model="claude-3-5-sonnet",
        stream=True,
    )

    # Apply the explicit context
    result = apply_context_overrides_to_call_args(
        call_args, context_override=explicit_context
    )

    # Check that the explicit context was applied
    assert result["provider"] == "anthropic"
    assert result["model"] == "claude-3-5-sonnet"
    assert result["stream"] is True

    # Test that it takes precedence over the current context
    with _context(
        provider="openai",
        model="gpt-4o",
        stream=False,
    ):
        # The explicit context should be used, not the current context
        result = apply_context_overrides_to_call_args(
            call_args, context_override=explicit_context
        )
        assert result["provider"] == "anthropic"
        assert result["model"] == "claude-3-5-sonnet"
        assert result["stream"] is True



"""Mirascope Model Context Protocol (MCP) implementation."""

from .client import MCPClient, sse_client, stdio_client
from .server import MCPServer  # DEPRECATED
from .tools import MCPTool  # DEPRECATED

__all__ = ["MCPClient", "MCPServer", "MCPTool", "sse_client", "stdio_client"]



from contextlib import suppress

from ._utils import configure

with suppress(ImportError):
    from ._with_hyperdx import with_hyperdx

from ._with_otel import with_otel

__all__ = ["configure", "with_hyperdx", "with_otel"]



"""Tests the `cohere.call_response` module."""

from cohere.types import (
    ApiMeta,
    ApiMetaBilledUnits,
    ChatMessage,
    NonStreamedChatResponse,
    ToolCall,
    ToolResult,
)

from mirascope.core import BaseMessageParam
from mirascope.core.cohere.call_response import CohereCallResponse
from mirascope.core.cohere.tool import CohereTool


def test_cohere_call_response() -> None:
    """Tests the `CohereCallResponse` class."""
    usage = ApiMetaBilledUnits(input_tokens=1, output_tokens=1)
    completion = NonStreamedChatResponse(
        generation_id="id",
        text="content",
        finish_reason="COMPLETE",
        meta=ApiMeta(billed_units=usage),
    )
    call_response = CohereCallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    call_response._model = "command-r-plus"
    assert call_response._provider == "cohere"
    assert call_response.content == "content"
    assert call_response.finish_reasons == ["COMPLETE"]
    assert call_response.model == "command-r-plus"
    assert call_response.id == "id"
    assert call_response.usage == usage
    assert call_response.input_tokens == 1
    assert call_response.output_tokens == 1
    assert call_response.cost == 1.8e-5
    assert call_response.message_param == ChatMessage(
        message="content",
        role="assistant",  # type: ignore
        tool_calls=None,
    )
    assert call_response.tools is None
    assert call_response.tool is None
    assert call_response.common_finish_reasons == ["stop"]
    assert call_response.common_message_param == BaseMessageParam(
        role="assistant", content="content"
    )
    assert call_response.common_user_message_param is None
    call_response.user_message_param = ChatMessage(
        message="content",
        role="user",  # type: ignore
    )
    assert call_response.common_user_message_param == BaseMessageParam(
        role="user", content="content"
    )


def test_cohere_call_response_with_tools() -> None:
    """Tests the `CohereCallResponse` class with tools."""

    class FormatBook(CohereTool):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    usage = ApiMetaBilledUnits(input_tokens=1, output_tokens=1)
    tool_call = ToolCall(
        name="FormatBook",
        parameters={"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
    )
    completion = NonStreamedChatResponse(
        generation_id="id",
        text="content",
        finish_reason="COMPLETE",
        meta=ApiMeta(billed_units=usage),
        tool_calls=[tool_call],
    )
    call_response = CohereCallResponse(
        metadata={},
        response=completion,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    call_response._model = "command-r-plus"
    assert call_response._provider == "cohere"
    assert call_response.content == "content"
    assert call_response.finish_reasons == ["COMPLETE"]
    assert call_response.model == "command-r-plus"
    assert call_response.id == "id"
    assert call_response.usage == usage
    assert call_response.input_tokens == 1
    assert call_response.output_tokens == 1
    assert call_response.cost == 1.8e-5
    assert call_response.message_param == ChatMessage(
        message="content",
        role="assistant",  # type: ignore
        tool_calls=[tool_call],
    )
    tools = call_response.tools
    assert tools and len(tools) == 1 and tools[0] == call_response.tool
    assert isinstance(call_response.tool, FormatBook)
    assert call_response.tool.title == "The Name of the Wind"
    assert call_response.tool.author == "Patrick Rothfuss"
    output = call_response.tool.call()
    assert output == "The Name of the Wind by Patrick Rothfuss"
    tool_message_params = call_response.tool_message_params(
        [(call_response.tool, output)]
    )
    assert tool_message_params == [
        ToolResult(call=tool_call, outputs=[{"output": output}])
    ]


def test_cohere_call_response_no_usage() -> None:
    """Tests the `CohereCallResponse` class with no usage."""
    completion = NonStreamedChatResponse(
        generation_id="id",
        text="content",
        finish_reason="COMPLETE",
        meta=None,
    )
    call_response = CohereCallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert call_response.usage is None
    assert call_response.input_tokens is None
    assert call_response.output_tokens is None



"""The `call_factory` method for generating provider specific call decorators."""

from collections.abc import AsyncIterable, Callable, Iterable
from enum import Enum
from functools import partial
from typing import TypeVar

from pydantic import BaseModel

from ._create import create_factory
from ._extract import extract_factory
from ._extract_with_tools import extract_with_tools_factory
from ._utils import (
    BaseType,
    GetJsonOutput,
    HandleStream,
    HandleStreamAsync,
    LLMFunctionDecorator,
    SameSyncAndAsyncClientSetupCall,
    SetupCall,
)
from ._utils._protocols import (
    AsyncLLMFunctionDecorator,
    CallDecorator,
    SyncLLMFunctionDecorator,
)
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .stream import BaseStream, stream_factory
from .stream_config import StreamConfig
from .structured_stream import structured_stream_factory
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType | Enum)
_ParsedOutputT = TypeVar("_ParsedOutputT")
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_AsyncBaseDynamicConfigT = TypeVar("_AsyncBaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseStreamT = TypeVar("_BaseStreamT", bound=BaseStream)
_SameSyncAndAsyncClientT = TypeVar("_SameSyncAndAsyncClientT", contravariant=True)
_SyncBaseClientT = TypeVar("_SyncBaseClientT", contravariant=True)
_AsyncBaseClientT = TypeVar("_AsyncBaseClientT", contravariant=True)
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_AsyncResponseT = TypeVar("_AsyncResponseT")
_AsyncResponseChunkT = TypeVar("_AsyncResponseChunkT")


def call_factory(  # noqa: ANN202
    *,
    TCallResponse: type[_BaseCallResponseT],
    TCallResponseChunk: type[_BaseCallResponseChunkT],
    TToolType: type[_BaseToolT],
    TStream: type[_BaseStreamT],
    default_call_params: BaseCallParams,
    setup_call: SameSyncAndAsyncClientSetupCall[
        _SameSyncAndAsyncClientT,
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _AsyncResponseT,
        _AsyncResponseChunkT,
        _BaseToolT,
    ]
    | SetupCall[
        _SyncBaseClientT,
        _AsyncBaseClientT,
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _AsyncResponseT,
        _AsyncResponseChunkT,
        _BaseToolT,
    ],
    get_json_output: GetJsonOutput[_BaseCallResponseT | _BaseCallResponseChunkT],
    handle_stream: HandleStream[_ResponseChunkT, _BaseCallResponseChunkT, _BaseToolT],
    handle_stream_async: HandleStreamAsync[
        _AsyncResponseChunkT, _BaseCallResponseChunkT, _BaseToolT
    ],
) -> CallDecorator[
    _BaseCallResponseT,
    _BaseCallResponseChunkT,
    _BaseDynamicConfigT,
    _AsyncBaseDynamicConfigT,
    _BaseCallParamsT,
    _BaseStreamT,
    _SyncBaseClientT,
    _AsyncBaseClientT,
    _SameSyncAndAsyncClientT,
]:
    """A factory method for creating provider-specific call decorators.

    Args:
        TCallResponse: The provider-specific `BaseCallResponse` type.
        TCallResponseChunk: The provider-specific `BaseCallResponseChunk` type.
        TToolType: The provider-specific `BaseTool` type.
        TStream: The provider-specific `BaseStream` type.
        default_call_params: The default call parameters to use, which must match the
            `TCallParams` type if provided.
        setup_call: The helper method for setting up a call, which returns the
            configured create function, the prompt template, the list of
            provider-specific messages, the list of provider-specific tool types, and
            the finalized `call_kwargs` with which to make the API call with the create
            function.
        get_json_output: The helper method for getting JSON output from a call response.
        handle_stream: The helper method for converting a provider's original stream
            generator into a generator that returns tuples of `(chunk, tool)` where
            `chunk` and `tool` are provider-specific `BaseCallResponseChunk` and
            `BaseTool` instances, respectively.
        handle_stream_async: The same helper method as `handle_stream` except for
            handling asynchronous streaming.
    """

    def base_call(
        model: str,
        *,
        stream: bool | StreamConfig = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT] | None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT]
        | Callable[[_BaseCallResponseChunkT], _ParsedOutputT]
        | Callable[[_ResponseModelT], _ParsedOutputT]
        | None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _AsyncBaseClientT
        | _SyncBaseClientT
        | None = None,
        call_params: BaseCallParams | None = None,
    ) -> (
        AsyncLLMFunctionDecorator[
            _AsyncBaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | BaseStream
            | _ResponseModelT
            | AsyncIterable[_ResponseModelT],
        ]
        | SyncLLMFunctionDecorator[
            _BaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | BaseStream
            | _ResponseModelT
            | Iterable[_ResponseModelT],
        ]
        | LLMFunctionDecorator[
            _BaseDynamicConfigT,
            _AsyncBaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | BaseStream
            | _ResponseModelT
            | Iterable[_ResponseModelT],
            _BaseCallResponseT
            | _ParsedOutputT
            | BaseStream
            | _ResponseModelT
            | AsyncIterable[_ResponseModelT],
        ]
    ):
        if stream and output_parser:
            raise ValueError("Cannot use `output_parser` with `stream=True`.")

        if call_params is None:
            call_params = default_call_params

        if response_model:
            if stream:
                return partial(
                    structured_stream_factory(
                        TCallResponse=TCallResponse,
                        TCallResponseChunk=TCallResponseChunk,
                        TStream=TStream,
                        TToolType=TToolType,
                        setup_call=setup_call,
                        get_json_output=get_json_output,
                    ),
                    model=model,
                    response_model=response_model,
                    json_mode=json_mode,
                    client=client,
                    call_params=call_params,
                )  # pyright: ignore [reportReturnType, reportCallIssue]
            elif tools:
                return partial(
                    extract_with_tools_factory(
                        TCallResponse=TCallResponse,
                        setup_call=setup_call,
                        get_json_output=get_json_output,
                    ),
                    model=model,
                    tools=tools,
                    response_model=response_model,
                    output_parser=output_parser,
                    client=client,
                    call_params=call_params,
                )  # pyright: ignore [reportReturnType, reportCallIssue]
            else:
                return partial(
                    extract_factory(
                        TCallResponse=TCallResponse,
                        TToolType=TToolType,
                        setup_call=setup_call,
                        get_json_output=get_json_output,
                    ),
                    model=model,
                    response_model=response_model,
                    output_parser=output_parser,
                    json_mode=json_mode,
                    client=client,
                    call_params=call_params,
                )  # pyright: ignore [reportCallIssue]

        if stream:
            return partial(
                stream_factory(
                    TCallResponse=TCallResponse,
                    TStream=TStream,
                    setup_call=setup_call,
                    handle_stream=handle_stream,
                    handle_stream_async=handle_stream_async,
                ),
                model=model,
                tools=tools,
                json_mode=json_mode,
                client=client,
                call_params=call_params,
                partial_tools=isinstance(stream, dict) and stream.get("partial_tools"),
            )  # pyright: ignore [reportReturnType, reportCallIssue]
        return partial(
            create_factory(
                TCallResponse=TCallResponse,
                setup_call=setup_call,
            ),
            model=model,
            tools=tools,
            response_model=None,
            output_parser=output_parser,
            json_mode=json_mode,
            client=client,
            call_params=call_params,
        )  # pyright: ignore [reportReturnType, reportCallIssue]

    return base_call  # pyright: ignore [reportReturnType]



"""The `XAITool` class for easy tool usage with xAI LLM calls.

usage docs: learn/tools.md
"""

from ..openai import OpenAITool


class XAITool(OpenAITool):
    """A simple wrapper around `OpenAITool`.

    Since xAI supports the OpenAI spec, we change nothing here.
    """



"""Tests the `bedrock.call_response` module."""

from typing import cast

import pytest
from mypy_boto3_bedrock_runtime.type_defs import (
    ConverseResponseTypeDef,
    MessageOutputTypeDef,
    TokenUsageTypeDef,
    ToolTypeDef,
)
from pydantic import ValidationError

from mirascope.core import BaseMessageParam
from mirascope.core.base import BaseCallKwargs
from mirascope.core.bedrock.call_params import BedrockCallParams
from mirascope.core.bedrock.call_response import BedrockCallResponse
from mirascope.core.bedrock.tool import BedrockTool


def test_bedrock_call_response() -> None:
    usage = TokenUsageTypeDef(inputTokens=1, outputTokens=1, totalTokens=2)
    message = MessageOutputTypeDef(content=[{"text": "content"}], role="assistant")
    response = ConverseResponseTypeDef(  # pyright: ignore [reportCallIssue]
        output={"message": message},
        ResponseMetadata={"RequestId": "id"},  # pyright: ignore [reportArgumentType]
        stopReason="end_turn",
        usage=usage,
        metrics={},  # pyright: ignore [reportArgumentType]
        trace={},
        additionalModelResponseFields={},
    )  # pyright: ignore [reportArgumentType]
    call_response = BedrockCallResponse(
        metadata={},
        response=response,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs=cast(
            BaseCallKwargs[ToolTypeDef],
            {"modelId": "anthropic.claude-3-haiku-20240307-v1:0"},
        ),
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert call_response._provider == "bedrock"
    assert call_response.content == "content"
    assert call_response.finish_reasons == ["end_turn"]
    assert call_response.model == "anthropic.claude-3-haiku-20240307-v1:0"
    assert call_response.id == "id"
    assert call_response.usage == usage
    assert call_response.input_tokens == 1
    assert call_response.output_tokens == 1
    assert call_response.cost is None
    assert call_response.message_param == {
        "role": "assistant",
        "content": [{"text": "content"}],
    }
    assert call_response.tools is None
    assert call_response.tool is None
    assert call_response.common_finish_reasons == ["stop"]
    assert call_response.common_message_param == BaseMessageParam(
        role="assistant", content="content"
    )
    assert call_response.common_user_message_param is None
    call_response.user_message_param = {
        "role": "user",
        "content": [{"text": "content"}],
    }
    assert call_response.common_user_message_param == BaseMessageParam(
        role="user", content="content"
    )


def test_bedrock_call_response_no_message() -> None:
    response = ConverseResponseTypeDef(  # pyright: ignore [reportCallIssue]
        output={},
        ResponseMetadata={"RequestId": "id"},
        stopReason="end_turn",
        usage=TokenUsageTypeDef(inputTokens=1, outputTokens=1),  # pyright: ignore [reportCallIssue]
    )
    call_response = BedrockCallResponse(
        metadata={},
        response=response,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={"modelId": "anthropic.claude-3-haiku-20240307-v1:0"},  # pyright: ignore [reportCallIssue, reportArgumentType]
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert call_response.content == ""
    assert call_response.message_param == {"role": "assistant", "content": []}
    assert call_response.tools is None


def test_bedrock_call_response_empty_content() -> None:
    message = MessageOutputTypeDef(content=[])  # pyright: ignore [reportCallIssue]
    response = ConverseResponseTypeDef(  # pyright: ignore [reportCallIssue]
        output={"message": message},
        ResponseMetadata={"RequestId": "id"},
        stopReason="end_turn",
        usage=TokenUsageTypeDef(inputTokens=1, outputTokens=1),  # pyright: ignore [reportCallIssue]
    )
    call_response = BedrockCallResponse(
        metadata={},
        response=response,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={"modelId": "anthropic.claude-3-haiku-20240307-v1:0"},  # pyright: ignore [reportCallIssue, reportArgumentType]
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert call_response.content == ""


def test_bedrock_call_response_with_tools() -> None:
    class FormatBook(BedrockTool):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    class AnotherTool(BedrockTool):
        @classmethod
        def _name(cls) -> str: ...

    tool_use = ToolTypeDef(  # pyright: ignore [reportCallIssue]
        toolUseId="id",
        input={"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
        name="FormatBook",
    )
    message = MessageOutputTypeDef(content=[{"toolUse": tool_use}, {"text": "content"}])  # pyright: ignore [reportCallIssue]
    response = ConverseResponseTypeDef(  # pyright: ignore [reportCallIssue]
        output={"message": message},
        ResponseMetadata={"RequestId": "id"},
        stopReason="end_turn",
        usage=TokenUsageTypeDef(inputTokens=1, outputTokens=2),  # pyright: ignore [reportCallIssue]
    )
    call_response = BedrockCallResponse(
        metadata={},
        response=response,
        tool_types=[FormatBook, AnotherTool],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={"modelId": "anthropic.claude-3-haiku-20240307-v1:0"},  # pyright: ignore [reportCallIssue, reportArgumentType]
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    tool = call_response.tool
    assert tools and len(tools) == 1 and tools[0] == tool
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    output = tool.call()
    assert output == "The Name of the Wind by Patrick Rothfuss"
    assert call_response.tool_message_params([(tool, output)]) == [
        {
            "role": "user",
            "content": [
                {
                    "toolResult": {
                        "content": [
                            {"text": "The Name of the Wind by Patrick Rothfuss"}
                        ],
                        "toolUseId": "id",
                    }
                }
            ],
        }
    ]


def test_bedrock_call_response_with_invalid_tool() -> None:
    class InvalidTool(BedrockTool):
        invalid_field: str

    tool_use = ToolTypeDef(  # pyright: ignore [reportCallIssue]
        toolUseId="id",
        input={"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
        name="InvalidTool",
    )
    message = MessageOutputTypeDef(content=[{"toolUse": tool_use}])  # pyright: ignore [reportCallIssue]
    response = ConverseResponseTypeDef(
        output={"message": message},
        ResponseMetadata={"RequestId": "id"},
        stopReason="end_turn",
        usage=TokenUsageTypeDef(inputTokens=1, outputTokens=2),  # pyright: ignore [reportCallIssue]
    )
    call_response = BedrockCallResponse(
        metadata={},
        response=response,
        tool_types=[InvalidTool],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={"modelId": "anthropic.claude-3-haiku-20240307-v1:0"},  # pyright: ignore [reportArgumentType]
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    with pytest.raises(ValidationError):
        _ = call_response.tools


def test_bedrock_call_response_with_multiple_tools() -> None:
    class FormatBook(BedrockTool):
        title: str
        author: str

        def call(self) -> str: ...

        @classmethod
        def _name(cls) -> str:
            return "FormatBook"

    class RecommendBook(BedrockTool):
        genre: str

        def call(self) -> str: ...

        @classmethod
        def _name(cls) -> str:
            return "RecommendBook"

    tool_use1 = ToolTypeDef(  # pyright: ignore [reportCallIssue]
        toolUseId="id1",
        input={"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
        name="FormatBook",
    )
    tool_use2 = ToolTypeDef(  # pyright: ignore [reportCallIssue]
        toolUseId="id2",
        input={"genre": "fantasy"},
        name="RecommendBook",
    )
    message = MessageOutputTypeDef(
        content=[  # pyright: ignore [reportCallIssue]
            {"toolUse": tool_use1},
            {"toolUse": tool_use2},
            {"text": "content"},
        ]
    )
    response = ConverseResponseTypeDef(  # pyright: ignore [reportCallIssue]
        output={"message": message},
        ResponseMetadata={"RequestId": "id"},
        stopReason="end_turn",
        usage=TokenUsageTypeDef(inputTokens=1, outputTokens=2),  # pyright: ignore [reportCallIssue]
    )
    call_response = BedrockCallResponse(
        metadata={},
        response=response,
        tool_types=[FormatBook, RecommendBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={"modelId": "anthropic.claude-3-haiku-20240307-v1:0"},  # pyright: ignore [reportArgumentType]
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    assert tools is not None
    assert len(tools) == 2
    assert isinstance(tools[0], FormatBook)
    assert isinstance(tools[1], RecommendBook)
    assert tools[0].title == "The Name of the Wind"
    assert tools[0].author == "Patrick Rothfuss"
    assert tools[1].genre == "fantasy"


def test_bedrock_call_response_with_mismatched_tool() -> None:
    class FormatBook(BedrockTool):
        title: str
        author: str

        def call(self) -> str: ...

        @classmethod
        def _name(cls) -> str:
            return "FormatBook"

    tool_use = ToolTypeDef(  # pyright: ignore [reportCallIssue]
        toolUseId="id",
        input={"genre": "fantasy"},
        name="RecommendBook",  # This doesn't match any tool type
    )
    message = MessageOutputTypeDef(content=[{"toolUse": tool_use}])  # pyright: ignore [reportCallIssue]
    response = ConverseResponseTypeDef(  # pyright: ignore [reportCallIssue]
        output={"message": message},
        ResponseMetadata={"RequestId": "id"},
        stopReason="end_turn",
        usage=TokenUsageTypeDef(inputTokens=1, outputTokens=2),  # pyright: ignore [reportCallIssue]
    )
    call_response = BedrockCallResponse(
        metadata={},
        response=response,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={"modelId": "anthropic.claude-3-haiku-20240307-v1:0"},  # pyright: ignore [reportArgumentType]
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    assert tools == []  # No tools should be extracted due to name mismatch


def test_bedrock_call_response_no_tool_types() -> None:
    tool_use = ToolTypeDef(  # pyright: ignore [reportCallIssue]
        toolUseId="id",
        input={"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
        name="FormatBook",
    )
    message = MessageOutputTypeDef(content=[{"toolUse": tool_use}])  # pyright: ignore [reportCallIssue]
    response = ConverseResponseTypeDef(  # pyright: ignore [reportCallIssue]
        output={"message": message},
        ResponseMetadata={"RequestId": "id"},
        stopReason="end_turn",
        usage=TokenUsageTypeDef(inputTokens=1, outputTokens=2),  # pyright: ignore [reportCallIssue]
    )
    call_response = BedrockCallResponse(
        metadata={},
        response=response,
        tool_types=None,  # No tool types provided
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={"modelId": "anthropic.claude-3-haiku-20240307-v1:0"},  # pyright: ignore [reportArgumentType]
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    assert tools is None  # Should return None when no tool types are provided



from ..core import CostMetadata, LocalProvider, Provider, calculate_cost
from ._call import call
from ._context import context
from ._override import override
from .call_response import CallResponse
from .call_response_chunk import CallResponseChunk
from .stream import Stream
from .tool import Tool

__all__ = [
    "CallResponse",
    "CallResponseChunk",
    "CostMetadata",
    "LocalProvider",
    "Provider",
    "Stream",
    "Tool",
    "calculate_cost",
    "call",
    "context",
    "override",
]
